package org.openxma.dsl.generator.mojo;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.eclipse.emf.mwe.core.WorkflowContext;
import org.eclipse.emf.mwe.core.WorkflowRunner;
import org.eclipse.emf.mwe.core.monitor.NullProgressMonitor;
import org.eclipse.emf.mwe.core.resources.CachingResourceLoaderImpl;
import org.eclipse.emf.mwe.core.resources.ResourceLoaderFactory;
import org.eclipse.emf.mwe.core.resources.ResourceLoaderImpl;
import org.openxma.dsl.generator.GeneratorProperties;

import at.spardat.xma.guidesign.util.ClassLoaderFactory;

public abstract class AbstractGeneratorMojo extends AbstractMojo {

    /**
     * The project itself. This parameter is set by maven.
     *
     * @parameter expression="${project}"
     * @required
     */
    protected MavenProject project;
    /**
     * @parameter default-value="${project.artifacts}"
     * @required
     * @readonly
     */
    private Set<Artifact> dependencies;

    /**
     * The location of the generator properties file.
     *
     * @parameter alias="properties"
     */
    private File generatorProperties;

    private List<URL> classLoaderUrls = new ArrayList<URL>();

    public void execute() throws MojoExecutionException, MojoFailureException {
        if (!Boolean.getBoolean("skipGenerator")) {
        	System.setProperty("org.eclipse.emf.ecore.EPackage.Registry.INSTANCE",
            "org.eclipse.emf.ecore.impl.EPackageRegistryImpl");
            doExecute();
        }
    }

    public abstract void doExecute() throws MojoExecutionException;

    protected WorkflowContext runWorkFlow(String workflowFile, Map<String, String> workflowParameters)
            throws MojoExecutionException {
        extendCurrentClassloader(getClass().getClassLoader());
        setUpGeneratorProperties(this.generatorProperties, workflowParameters);
        WorkflowRunner workflowRunner = new WorkflowRunner();
        HashMap<String, Object> externalSlotContents = getExternalSlotContents();
		if (!workflowRunner.run(workflowFile, new NullProgressMonitor(), workflowParameters,
                externalSlotContents)) {
            throw new MojoExecutionException("Errors while executing workflow '" + workflowFile + "'");
        }
        return workflowRunner.getContext();
    }

    protected HashMap<String, Object> getExternalSlotContents() {
		return new HashMap<String, Object>();
	}

    private void setUpGeneratorProperties(File propertiesFile, Map<String, String> workflowParameters) throws MojoExecutionException {
        Properties properties = new Properties();
		if (propertiesFile != null) {
			getLog().info("load generator properties from '" + propertiesFile.getPath() + "'");
			try {
				properties.load(new FileInputStream(propertiesFile));
			} catch (FileNotFoundException e) {
				throw new MojoExecutionException(e.getMessage(), e);
			} catch (IOException e) {
				throw new MojoExecutionException(e.getMessage(), e);
			}
		}
		putWorkflowParametersToGeneratorProperties(properties,workflowParameters);
		GeneratorProperties.setProperties(properties);
	}

    /**
     * Method gives {@link DomGeneratorMojo} a chance to put additional
     * properties into {@link GeneratorProperties}. </br> Up until now
     * {@link PomGeneratorMojo} has no need to do so as well.
     *
     * @param props
     * @param workflowParameters
     * @return
     */
    public void putWorkflowParametersToGeneratorProperties(Properties props,
            Map<String, String> workflowParameters) {}

    protected void extendCurrentClassloader(ClassLoader currentClassLoader) {
        if (classLoaderUrls.isEmpty()) {
            classLoaderUrls.add(toURL(resolvePath(new File(project.getBuild().getSourceDirectory())), false));
            List<?> resources = project.getBuild().getResources();
            if (resources != null) {
                for (int i = 0; i < resources.size(); i++) {
                    org.apache.maven.model.Resource resource = (org.apache.maven.model.Resource) resources.get(i);
                    File directory = resolvePath(new File(resource.getDirectory()));
                    classLoaderUrls.add(toURL(directory, false));
                }
            }
            
            for (Artifact artifact : dependencies) {
                URL url = toURL(artifact.getFile(), false);
                classLoaderUrls.add(url);
            }

            // TODO: consider active maven profiles

            if (classLoaderUrls.size() > 0) {
                File outputDirectory = resolvePath(new File(project.getBuild().getOutputDirectory()));
                classLoaderUrls.add(toURL(outputDirectory, false));
                URL[] urls = classLoaderUrls.toArray(new URL[classLoaderUrls.size()]);
                ClassLoader urlClassLoader = new URLClassLoader(urls, currentClassLoader);
                ResourceLoaderFactory.setCurrentThreadResourceLoader(new CachingResourceLoaderImpl(
                        new ResourceLoaderImpl(urlClassLoader)));
                // required for guidesigner!
                ClassLoaderFactory.setProjectClassLoader(urlClassLoader);
                // required for backwards compat. with Pom2Xma.mwe can be
                // removed if we change and remove the classpathUri to 0fileUri
                Thread.currentThread().setContextClassLoader(urlClassLoader);
                String separator = System.getProperty("path.separator");
                StringBuilder stringBuilder = new StringBuilder();
                for (URL url : urls) {
                    File file = new File(url.getPath());
                    if (file.exists() && file.isDirectory()
                            || (file.getName().endsWith(".jar") || file.getName().endsWith(".zip"))) {
                        stringBuilder.append(url.getPath());
                        stringBuilder.append(separator);
                    }
                }
                System.setProperty("java.class.path", stringBuilder.toString());
            }
        }
    }

    private URL toURL(File path, boolean trailingSlash) {
        URL url = null;
        try {
            url = new URL(path.toURI().toString() + ((trailingSlash) ? "/" : ""));
        } catch (Exception e) {
            getLog().error("Could not resolve \"" + path.getPath() + "\".", e);
        }
        return url;
    }

    private File resolvePath(File path) {
        if (!path.isAbsolute()) {
            path = new File(project.getBasedir(), path.getPath());
        }
        if (path.isDirectory()) {
            String name = path.getName();
            if (!name.endsWith(File.separator)) {
                path = new File(path.getPath() + File.separator);
            }
        }
        return path;
    }

    protected String join(Object[] array, String separator, int startIndex, int endIndex) {
        if (array == null) {
            return null;
        }
        if (separator == null) {
            separator = "";
        }

        int bufSize = (endIndex - startIndex);
        if (bufSize <= 0) {
            return "";
        }

        bufSize *= ((array[startIndex] == null ? 16 : array[startIndex].toString().length()) + separator.length());

        StringBuffer buf = new StringBuffer(bufSize);

        for (int i = startIndex; i < endIndex; i++) {
            if (i > startIndex) {
                buf.append(separator);
            }
            if (array[i] != null) {
                buf.append(array[i]);
            }
        }
        return buf.toString();
    }

}
